==> ciao-4.6-src-core.tar.gz: ciao-4.6/src/lib/datamodel/ftkernel/ftfilter.c <==
/*                                                                
**  Copyright (C) 2002-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   ftfilter.c                                                  */
/*                                                               */
/*   Functions that access blocks                                */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmilib.h"
#include "ftkernelP.h"


/***************************************************************/
/*                                                             */
/*  Function:   ftp_build_cfitsio_spec                         */
/*                                                             */
/*  Description:                                               */
/*     Here we build a CFITSIO filter if the run time subspace */
/*     is very simple                                          */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock  *block       - block                  */
/*              dmDescriptor  *rts_dd - subspace descriptor    */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                  void                                       */
/*                                                             */
/***************************************************************/
void ftp_build_cfitsio_spec( dmBlock* block, dmDescriptor* rts_dd )
{
  dmDescriptor *dd     = NULL;
  dmDescriptor *ddcpt  = NULL;
  dmRangeList  *range  = NULL;
  dmFilter     *filter = NULL;
  char         *axisname = NULL;
  double vals[2]; 
  long vcpt = 0;
  long dim  = 0;
  int  incnull = 0;

  dd = dmi_descriptor_get_parent( rts_dd );
 
  if (dmi_block_get_raw_type( block ) == dmIMAGE )
    return;

  /* Ignore row filter */
  axisname = dmi_descriptor_get_name( dd );

  if (cs_uceq( axisname, "#row" ))
    return;

  /* If complicated, bail!  Only called if ncpts = 1 */
  dim = dmi_descriptor_get_element_dim( rts_dd );

  for  ( vcpt = 1; vcpt <= dim; vcpt++ )
  {
    ddcpt = dmi_descriptor_get_cpt( rts_dd, vcpt );
    axisname = dmi_descriptor_get_name( ddcpt );

    if ( dmi_filter_get_simple_range( rts_dd, vcpt, vals ))
    {
      filter = dmi_subspace_col_get_filter( rts_dd, 1 );
      range = dmi_filter_get_range( filter, vcpt );
      incnull = dmi_rangelist_get_incnull( range );
      ftp_fits_range_spec( block, axisname, vals[0], vals[1], incnull );
    }
  }
}


/***************************************************************/
/*                                                             */
/*  Function:   ftp_fits_range_spec                            */
/*                                                             */
/*  Description:                                               */
/*      Construct a CFITSO range pre-filter                    */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock  *block       - block                  */
/*              char     *name        -                        */
/*              double    tmin        -                        */
/*              double    tmax        -                        */
/*              int       incnull     -                        */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                  void                                       */
/*                                                             */
/***************************************************************/
void ftp_fits_range_spec( dmBlock* block, char* name, double tmin, double tmax, int incnull )
{
  ftBlock *ftblock = NULL;

  char  tmp[200];
  char  tmp1[80];
  char *ptr = tmp;
  char *filter;
  long  nf = 0;
  long  nfs = 0;


  if ( tmin > -DT_REAL8_MAX )
  {
    dmu_print_number( tmin  - fabs( tmin ) * DM_FILT_EPSILON, tmp1, CARD_SIZE );
    ptr += sprintf( ptr, "%s>=%s ", name, tmp1 );
    ++nf;
  }

  if ( tmax < DT_REAL8_MAX )
  {
    dmu_print_number( tmax  + fabs( tmax ) * DM_FILT_EPSILON, tmp1, CARD_SIZE );

    if (nf > 0 )
      ptr += sprintf( ptr, " && " );
    sprintf( ptr, "%s<=%s", name, tmp1 );
  }

  if ( incnull == 1 )
  {
    ptr += sprintf( ptr, "|| ISNULL(%s)", name );
  }
 
  ftblock = (ftBlock*)dmi_block_get_kb( block );

  if ( !ftblock->filter )
  {
    filter = cs_dup( tmp );
  }
  else
  {  
    nfs = strlen(ftblock->filter ) + strlen( tmp ) + 4;
    filter = realloc( ftblock->filter, nfs );
    strcat( filter, "&& " );
    strcat( filter, tmp );
  }
  ftblock->filter = filter;
}


/***************************************************************/
/*                                                             */
/*  Function:   ftp_block_reopen_with_filter                   */
/*                                                             */
/*  Description:                                               */
/*      Reopen the file with a CFITSIO filter applied in the   */
/*      hope that this will improve efficiency.                */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock  *block       - block                  */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                  void                                       */
/*                                                             */
/***************************************************************/
void ftp_block_reopen_with_filter( dmBlock* block )
{
  ftDataset *ftds = NULL;
  fitsfile  *fptr = NULL;
  dmDataset *ds   = NULL;
  ftBlock   *ftblock = NULL;

  char *dsname = NULL;
  char *name   = NULL;

  long nmegabytes = 0;
  long maxmeg     = 0; 
  long m          = 0;
  long nrows      = 0;
  int  ftstatus = FITS_KERNEL_OK;


  /* 
   * Swap the FITS file out from under the user.
   * We reopen the file with a CFITSIO filter applied in the hope that
   * this will improve efficiency. 
   * Note that this will invalidate the #row descriptor.
   */

  ftblock = (ftBlock*)dmi_block_get_kb( block );
 
  if ( ftblock && ftblock->filter )
  {
    nmegabytes = (long)(ftblock->bytes / 1000000.0);
    maxmeg = dmi_block_get_maxmeg( block );

    if( nmegabytes > maxmeg ) /* Too big, FITS would steal all our memory */
      return;

    ds = (dmDataset*)dmi_block_get_ds( block );
    ftds = (ftDataset*)dmi_dataset_get_kds( ds );
    dsname = dmi_dataset_get_name( ds );
    m = strlen( ftblock->filter ) + strlen( dsname ) + 20;
    name = malloc( m );
    sprintf( name, "%s[%ld][%s]", dsname, dmi_block_get_no( block )-1, ftblock->filter );

    if ( dmverbose >= 1 )
      fprintf( stderr, "CFITSIO label: %s\n", name );

    fits_open_file(&fptr,name,READONLY,&ftstatus); /* CFITSIO */

    if ( ftstatus != 0 )
    {
      if ( dmverbose >= 4 )
        fprintf( stderr, "Warning: FITS error %d on FITS reopen %s\n", ftstatus, name );
      return;
    }

    fits_close_file(ftds->fptr,&ftstatus);  /* CFITSIO */
    ftds->fptr = fptr;
    fits_read_key( fptr, TLONG, "NAXIS2", &nrows, NULL, &ftstatus ); /* CFITSIO */
    dmi_block_set_nrows( block, nrows );

    if ( ftstatus != 0 )
      fprintf( stderr, "Error %d on FITS reopen %s\n", ftstatus, name );
    free( name );
  }
}


#if 0
/*
 *  Construct a CFITSO ISNULL pre-filter
 */
void ftp_fits_null_spec( dmBlock* block, char* name )
{
 char tmp[200];
 ftBlock* ftblock;
 char* ptr = tmp;
 char* filter;
 long nfs;
 long nf;

 ptr += sprintf( ptr, "ISNULL(%s)", name );
 nf = 1;

 ftblock = (ftBlock*)dmi_block_get_kb( block );
 if ( !ftblock->filter )
 {
  filter = cs_dup( tmp );
 } else {  
  nfs = strlen(ftblock->filter ) + strlen( tmp ) + 4;
  filter = realloc( ftblock->filter, nfs );
  strcat( filter, "&& " );
  strcat( filter, tmp );
 }
 ftblock->filter = filter;
}
#endif

==> ciao-4.8-src-core.tar.gz: ciao-4.8/src/lib/datamodel/ftkernel/ftfilter.c <==
/*                                                                
**  Copyright (C) 2002-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   ftfilter.c                                                  */
/*                                                               */
/*   Functions that access blocks                                */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmilib.h"
#include "ftkernelP.h"


/***************************************************************/
/*                                                             */
/*  Function:   ftp_build_cfitsio_spec                         */
/*                                                             */
/*  Description:                                               */
/*     Here we build a CFITSIO filter if the run time subspace */
/*     is very simple                                          */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock  *block       - block                  */
/*              dmDescriptor  *rts_dd - subspace descriptor    */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                  void                                       */
/*                                                             */
/***************************************************************/
void ftp_build_cfitsio_spec( dmBlock* block, dmDescriptor* rts_dd )
{
  dmDescriptor *dd     = NULL;   /* parent (column) descriptor */
  dmDescriptor *ddcpt  = NULL;   /* component descriptor       */
  dmRangeList  *range  = NULL;   /* filter ranges              */
  dmFilter     *filter = NULL;   /* filter definition          */
  char         *axisname = NULL; /* descriptor name            */
  double vals[2];                /* simple range values        */
  long vcpt = 0;                 /* vector component number    */
  long dim  = 0;                 /* vector dimensionality      */
  int  incnull = 0;              /* filter includes nulls?     */
  dmDataType dtype;              /* descriptor datatype        */
  double nulval;                 /* null value on descriptor   */
  dmBool nullset;                /* null is set?               */

  dd = dmi_descriptor_get_parent( rts_dd );
  dtype = dmi_descriptor_get_datatype( dd );
  nullset = dmi_descriptor_get_null( dd, (void*)&nulval );

  /* Ignore images */
  if (dmi_block_get_raw_type( block ) == dmIMAGE )
    return;

  /* Ignore row filter */
  axisname = dmi_descriptor_get_name( dd );
  if (cs_uceq( axisname, "#row" ))
    return;

  /* Ignore boolean columns (numeric filter doesn't work in cfitsio) */
  if ( dtype == DT_LOGICAL )
    return;
  
  /* If complicated, bail!  Only called if ncpts = 1 */

  /* Process vector components separately */
  dim = dmi_descriptor_get_element_dim( rts_dd );
  for  ( vcpt = 1; vcpt <= dim; vcpt++ )
  {
    ddcpt = dmi_descriptor_get_cpt( rts_dd, vcpt );
    axisname = dmi_descriptor_get_name( ddcpt );

    if ( dmi_filter_get_simple_range( rts_dd, vcpt, vals ) )
    {
      filter = dmi_subspace_col_get_filter( rts_dd, 1 );
      range = dmi_filter_get_range( filter, vcpt );
      incnull = dmi_rangelist_get_incnull( range );

      if ( incnull && nullset && ((dtype == DT_REAL4)||(dtype == DT_REAL8)) )
      {
	/* Floating point columns with null value on descriptor comes    */
	/* from TDNULL.  Skip pre-filter so that the DM can handle them. */
	continue;
      }
      else
        ftp_fits_range_spec( block, axisname, vals[0], vals[1], incnull );
    }
  }
}


/***************************************************************/
/*                                                             */
/*  Function:   ftp_fits_range_spec                            */
/*                                                             */
/*  Description:                                               */
/*      Construct a CFITSO range pre-filter                    */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock  *block       - block                  */
/*              char     *name        -                        */
/*              double    tmin        -                        */
/*              double    tmax        -                        */
/*              int       incnull     -                        */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                  void                                       */
/*                                                             */
/***************************************************************/
void ftp_fits_range_spec( dmBlock* block, char* name, double tmin, double tmax, int incnull )
{
  ftBlock *ftblock = NULL;

  char  tmp[200];
  char  tmp1[80];
  char *ptr = tmp;
  char *filter;
  long  nf = 0;
  long  nfs = 0;


  if ( tmin > -DT_REAL8_MAX )
  {
    dmu_print_number( tmin  - fabs( tmin ) * DM_FILT_EPSILON, tmp1, CARD_SIZE );
    ptr += sprintf( ptr, "%s>=%s ", name, tmp1 );
    ++nf;
  }

  if ( tmax < DT_REAL8_MAX )
  {
    dmu_print_number( tmax  + fabs( tmax ) * DM_FILT_EPSILON, tmp1, CARD_SIZE );

    if (nf > 0 )
      ptr += sprintf( ptr, " && " );
      ptr += sprintf( ptr, "%s<=%s", name, tmp1 );
  }

  if ( incnull == 1 )
  {
    ptr += sprintf( ptr, " || ISNULL(%s)", name );
  }
 
  ftblock = (ftBlock*)dmi_block_get_kb( block );

  if ( !ftblock->filter )
  {
    filter = cs_dup( tmp );
  }
  else
  {  
    nfs = strlen(ftblock->filter ) + strlen( tmp ) + 4;
    filter = realloc( ftblock->filter, nfs );
    strcat( filter, "&& " );
    strcat( filter, tmp );
  }
  ftblock->filter = filter;
}


/***************************************************************/
/*                                                             */
/*  Function:   ftp_block_reopen_with_filter                   */
/*                                                             */
/*  Description:                                               */
/*      Reopen the file with a CFITSIO filter applied in the   */
/*      hope that this will improve efficiency.                */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock  *block       - block                  */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                  void                                       */
/*                                                             */
/***************************************************************/
void ftp_block_reopen_with_filter( dmBlock* block )
{
  ftDataset *ftds = NULL;
  fitsfile  *fptr = NULL;
  dmDataset *ds   = NULL;
  ftBlock   *ftblock = NULL;

  char *dsname = NULL;
  char *name   = NULL;

  long nmegabytes = 0;
  long maxmeg     = 0; 
  long m          = 0;
  long nrows      = 0;
  int  ftstatus = FITS_KERNEL_OK;


  /* 
   * Swap the FITS file out from under the user.
   * We reopen the file with a CFITSIO filter applied in the hope that
   * this will improve efficiency. 
   * Note that this will invalidate the #row descriptor.
   */

  ftblock = (ftBlock*)dmi_block_get_kb( block );
 
  if ( ftblock && ftblock->filter )
  {
    nmegabytes = (long)(ftblock->bytes / 1000000.0);
    maxmeg = dmi_block_get_maxmeg( block );

    if( nmegabytes > maxmeg ) /* Too big, FITS would steal all our memory */
      return;

    ds = (dmDataset*)dmi_block_get_ds( block );
    ftds = (ftDataset*)dmi_dataset_get_kds( ds );
    dsname = dmi_dataset_get_name( ds );
    m = strlen( ftblock->filter ) + strlen( dsname ) + 20;
    name = malloc( m );
    sprintf( name, "%s[%ld][%s]", dsname, dmi_block_get_no( block )-1, ftblock->filter );

    if ( dmverbose >= 1 )
      fprintf( stderr, "CFITSIO label: %s\n", name );

    fits_open_file(&fptr,name,READONLY,&ftstatus); /* CFITSIO */

    if ( ftstatus != 0 )
    {
      if ( dmverbose >= 4 )
        fprintf( stderr, "Warning: FITS error %d on FITS reopen %s\n", ftstatus, name );
      return;
    }

    fits_close_file(ftds->fptr,&ftstatus);  /* CFITSIO */
    ftds->fptr = fptr;
    fits_read_key( fptr, TLONG, "NAXIS2", &nrows, NULL, &ftstatus ); /* CFITSIO */
    dmi_block_set_nrows( block, nrows );

    if ( ftstatus != 0 )
      fprintf( stderr, "Error %d on FITS reopen %s\n", ftstatus, name );
    free( name );
  }
}


#if 0
/*
 *  Construct a CFITSO ISNULL pre-filter
 */
void ftp_fits_null_spec( dmBlock* block, char* name )
{
 char tmp[200];
 ftBlock* ftblock;
 char* ptr = tmp;
 char* filter;
 long nfs;
 long nf;

 ptr += sprintf( ptr, "ISNULL(%s)", name );
 nf = 1;

 ftblock = (ftBlock*)dmi_block_get_kb( block );
 if ( !ftblock->filter )
 {
  filter = cs_dup( tmp );
 } else {  
  nfs = strlen(ftblock->filter ) + strlen( tmp ) + 4;
  filter = realloc( ftblock->filter, nfs );
  strcat( filter, "&& " );
  strcat( filter, tmp );
 }
 ftblock->filter = filter;
}
#endif
